@@ -11,7 +11,6 @@ LANG=C
 SCRIPT_DIR=$(dirname $(readlink -f $0))
 TOP=$(readlink -f $SCRIPT_DIR/../)
 TEST_DEV=${TEST_DEV:-}
-TEST_MNT=${TEST_MNT:-$TOP/tests/mnt}
 RESULTS="$TOP/tests/fsck-tests-results.txt"
 
 source $TOP/tests/common
@@ -20,17 +19,35 @@ source $TOP/tests/common
 export TOP
 export RESULTS
 # For custom script needs to verfiy recovery
-export TEST_MNT
 export LANG
 
 rm -f $RESULTS
-mkdir -p $TEST_MNT || _fail "unable to create mount point on $TEST_MNT"
 
 # test rely on corrupting blocks tool
 check_prereq btrfs-corrupt-block
 check_prereq btrfs-image
 check_prereq btrfs
 
+run_one_test() {
+	local testname
+
+	testname="$1"
+	echo "    [TEST]   $(basename $testname)"
+	cd $testname
+	echo "=== Entering $testname" >> $RESULTS
+	if [ -x test.sh ]; then
+		# Type 2
+		./test.sh
+		if [ $? -ne 0 ]; then
+			_fail "test failed for case $(basename $testname)"
+		fi
+	else
+		# Type 1
+		check_all_images `pwd`
+	fi
+	cd $TOP
+}
+
 # Each dir contains one type of error for btrfsck test.
 # Each dir must be one of the following 2 types:
 # 1) Only btrfs-image dump
@@ -44,20 +61,8 @@ check_prereq btrfs
 #    This is for case btrfs-image can't dump or case needs extra
 #    check/verify
 
-for i in $(find $TOP/tests/fsck-tests -maxdepth 1 -mindepth 1 -type d | sort)
+for i in $(find $TOP/tests/fsck-tests -maxdepth 1 -mindepth 1 -type d	\
+	${TEST:+-name "$TEST"} | sort)
 do
-	echo "    [TEST]   $(basename $i)"
-	cd $i
-	echo "=== Entering $i" >> $RESULTS
-	if [ -x test.sh ]; then
-		# Type 2
-		./test.sh
-		if [ $? -ne 0 ]; then
-			_fail "test failed for case $(basename $i)"
-		fi
-	else
-		# Type 1
-		check_all_images `pwd`
-	fi
-	cd $TOP
+	run_one_test "$i"
 done
